/*
 * Copyright (C) 2018 Red Hat, Inc. (www.redhat.com)
 *
 * This library is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

#include "evolution-data-server-config.h"

#include <glib/gi18n-lib.h>

#include "e-oauth2-service.h"
#include "e-oauth2-service-base.h"

#include "e-oauth2-service-yandex.h"

/* Forward Declarations */
static void e_oauth2_service_yandex_oauth2_service_init (EOAuth2ServiceInterface *iface);

G_DEFINE_TYPE_WITH_CODE (EOAuth2ServiceYandex, e_oauth2_service_yandex, E_TYPE_OAUTH2_SERVICE_BASE,
	G_IMPLEMENT_INTERFACE (E_TYPE_OAUTH2_SERVICE, e_oauth2_service_yandex_oauth2_service_init))

static gboolean
eos_yandex_guess_can_process (EOAuth2Service *service,
			      const gchar *protocol,
			      const gchar *hostname)
{
	return hostname && (
		e_util_utf8_strstrcase (hostname, ".yandex.ru") ||
		e_util_utf8_strstrcase (hostname, ".ya.ru") ||
		e_util_utf8_strstrcase (hostname, ".yandex.com"));
}

static const gchar *
eos_yandex_get_name (EOAuth2Service *service)
{
	return "Yandex";
}

static const gchar *
eos_yandex_get_display_name (EOAuth2Service *service)
{
	/* Translators: This is a user-visible string, display name of an OAuth2 service. */
	return C_("OAuth2Service", "yandex");
}

static const gchar *
eos_yandex_read_settings (EOAuth2Service *service,
			  const gchar *key_name)
{
	G_LOCK_DEFINE_STATIC (user_settings);
	gchar *value;

	G_LOCK (user_settings);

	value = g_object_get_data (G_OBJECT (service), key_name);
	if (!value) {
		GSettings *settings;

		settings = g_settings_new ("org.gnome.evolution-data-server");
		value = g_settings_get_string (settings, key_name);
		g_object_unref (settings);

		if (value && *value) {
			g_object_set_data_full (G_OBJECT (service), key_name, value, g_free);
		} else {
			g_free (value);
			value = (gchar *) "";

			g_object_set_data (G_OBJECT (service), key_name, value);
		}
	}

	G_UNLOCK (user_settings);

	return value;
}

static const gchar *
eos_yandex_get_client_id (EOAuth2Service *service,
			  ESource *source)
{
	const gchar *client_id;

	client_id = eos_yandex_read_settings (service, "oauth2-yandex-client-id");

	if (client_id && *client_id)
		return client_id;

	return YANDEX_CLIENT_ID;
}

static const gchar *
eos_yandex_get_client_secret (EOAuth2Service *service,
			      ESource *source)
{
	const gchar *client_secret;

	client_secret = eos_yandex_read_settings (service, "oauth2-yandex-client-secret");

	if (client_secret && *client_secret)
		return client_secret;

	return YANDEX_CLIENT_SECRET;
}

static const gchar *
eos_yandex_get_authentication_uri (EOAuth2Service *service,
				   ESource *source)
{
	return "https://oauth.yandex.ru/authorize";
}

static const gchar *
eos_yandex_get_refresh_uri (EOAuth2Service *service,
			    ESource *source)
{
	return "https://oauth.yandex.ru/token";
}

static const gchar *
eos_yandex_get_redirect_uri (EOAuth2Service *service,
			    ESource *source)
{
	return "http://localhost";
}

static void
eos_yandex_prepare_authentication_uri_query (EOAuth2Service *service,
					     ESource *source,
					     GHashTable *uri_query)
{
	const gchar *YANDEX_SCOPE =
		/* GMail IMAP and SMTP access */
		"login:email login:info mail:imap_full mail:imap_ro mail:smtp calendar:all";

	g_return_if_fail (uri_query != NULL);

	e_oauth2_service_util_set_to_form (uri_query, "scope", YANDEX_SCOPE);
}

static gboolean
eos_yandex_extract_authorization_code (EOAuth2Service *service,
				       ESource *source,
				       const gchar *page_title,
				       const gchar *page_uri,
				       const gchar *page_content,
				       gchar **out_authorization_code)
{
	g_return_val_if_fail (out_authorization_code != NULL, FALSE);

	*out_authorization_code = NULL;

	if (page_uri && *page_uri) {
		SoupURI *suri;

		suri = soup_uri_new (page_uri);
		if (suri) {
			const gchar *query = soup_uri_get_query (suri);
			gboolean known = FALSE;

			if (query && *query) {
				GHashTable *params;

				params = soup_form_decode (query);
				if (params) {
					const gchar *response;

					response = g_hash_table_lookup (params, "code");
					if (response) {
						*out_authorization_code = g_strdup (response);
						known = TRUE;
					}

					g_hash_table_destroy (params);
				}
			}

			soup_uri_free (suri);

			if (known)
				return TRUE;
		}
	}

	return FALSE;
}

static void
e_oauth2_service_yandex_oauth2_service_init (EOAuth2ServiceInterface *iface)
{
	iface->guess_can_process = eos_yandex_guess_can_process;
	iface->get_name = eos_yandex_get_name;
	iface->get_display_name = eos_yandex_get_display_name;
	iface->get_client_id = eos_yandex_get_client_id;
	iface->get_client_secret = eos_yandex_get_client_secret;
	iface->get_authentication_uri = eos_yandex_get_authentication_uri;
	iface->get_refresh_uri = eos_yandex_get_refresh_uri;
	iface->get_redirect_uri = eos_yandex_get_redirect_uri;
	iface->prepare_authentication_uri_query = eos_yandex_prepare_authentication_uri_query;
	iface->extract_authorization_code = eos_yandex_extract_authorization_code;
}

static void
e_oauth2_service_yandex_class_init (EOAuth2ServiceYandexClass *klass)
{
}

static void
e_oauth2_service_yandex_init (EOAuth2ServiceYandex *oauth2_google)
{
}
